
/* wrightFlightServer - This software is licenced under the GPL licence:
   read the file LICENCE for more information about the licence. */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include "defines.h"
#include "main.h"
#include "game.h"
#include "stats.h"


// the server port
#define SERVER_PORT 8889

// the size of the input buffer (64KBs should be enough for UDP packets)
#define INPUT_BUFFER_SIZE (64*1024)

// debug to stderr?
#define DEBUG_MESSAGES

// the one and only game structure
struct game game;

// quit?
int quit = NO;

// the client timeout in seconds
int clientTimeoutSeconds;


void clientReset(struct game *game, int clientID) {

	struct client *client = &game->clients[clientID];

	client->iP = 0;
	client->packetID = 0;
	client->alive = NO;
	client->ready = NO;
	client->kills = 0;
	client->stats = NULL;
	client->planeModel = 0;
	client->colorR = 15;
	client->colorG = 15;
	client->colorB = 15;
	client->statisticsScrollY = 0;
	client->lastPingMS = 0;
	client->pingsSent = 0;
}


void gameReset(struct game *game) {

	int i;

	// init the game
	game->mode = GAME_MODE_TAKING_CLIENTS;
	game->clientsN = 0;
	game->level = 0;
	game->playMode = GAME_MULTIPLAYER_MODE_3_LIVES;
	game->stateBits = 0;

	// zero the clients
	for (i = 0; i < CLIENTS_MAX; i++)
		clientReset(game, i);
}


void clientQuit(struct game *game, int clientID) {

	// kill the client, but keep the name and statistics
	game->clients[clientID].alive = NO;
	game->clients[clientID].iP = 0;

	// decrease the count
	game->clientsN--;

	if (game->clientsN <= 0) {
		// end of game? all quit?

#if defined(DEBUG_MESSAGES)
		fprintf(stderr, "CLIENTQUIT: All clients have quit... Resetting the server...\n");
#endif

		// reset the game (no clients, etc.)
		gameReset(game);
	}
}


void gameSendStatistics(struct game *game, struct sockaddr_in *addrClient, unsigned char *outputBuffer, int sockFD, int clientID) {

	struct client *client;
	struct mac *mac;
	struct vs *vs;
	int i, j, k, m;

	client = &(game->clients[clientID]);

	// no statistics to send?
	if (client->stats == NULL)
		return;

	vs = client->stats->vs;

	// skip the statisticsScrollY entries
	for (i = 0; i < client->statisticsScrollY; i++) {
		if (vs != NULL)
			vs = vs->next;
	}

	// output 
	j = 2;
	outputBuffer[j++] = 'l';

	// plays
	i = client->stats->plays;
	outputBuffer[j++] = i >> 24;
	outputBuffer[j++] = i >> 16;
	outputBuffer[j++] = i >> 8;
	outputBuffer[j++] = i;

	// the eight entries
	for (k = 0; k < 8; k++) {
		if (vs != NULL) {
			// name
			mac = statsMACGet(game, vs->macAddress);
			if (mac == NULL) {
				outputBuffer[j++] = 1;
				outputBuffer[j++] = '?';
			}
			else {
				i = strlen(mac->name);
				outputBuffer[j++] = i;

				for (m = 0; m < i; m++)
					outputBuffer[j++] = mac->name[m];
			}

			// wins
			i = vs->wins;
			outputBuffer[j++] = i >> 24;
			outputBuffer[j++] = i >> 16;
			outputBuffer[j++] = i >> 8;
			outputBuffer[j++] = i;

			// losses
			i = vs->losses;
			outputBuffer[j++] = i >> 24;
			outputBuffer[j++] = i >> 16;
			outputBuffer[j++] = i >> 8;
			outputBuffer[j++] = i;

			vs = vs->next;
		}
		else {
			// name
			outputBuffer[j++] = 0;

			// wins
			outputBuffer[j++] = 0;
			outputBuffer[j++] = 0;
			outputBuffer[j++] = 0;
			outputBuffer[j++] = 0;

			// losses
			outputBuffer[j++] = 0;
			outputBuffer[j++] = 0;
			outputBuffer[j++] = 0;
			outputBuffer[j++] = 0;
		}
	}

	// set the size
	outputBuffer[0] = j >> 8;
	outputBuffer[1] = j & 0xFF;

	// fix the structure
	addrClient->sin_port = client->port;
	addrClient->sin_addr.s_addr = client->iP;

	// send
	if (sendto(sockFD, outputBuffer, j, 0, (struct sockaddr *)addrClient, sizeof(struct sockaddr)) == -1)
		fprintf(stderr, "GAMESENDSTATISTICS: [1] Error sending a message.\n");
}


void gameDistribute(struct game *game, struct sockaddr_in *addrClient, unsigned char *outputBuffer, int sockFD, int size) {

	int i;

	// write the size
	outputBuffer[0] = size >> 8;
	outputBuffer[1] = size & 0xFF;

	// pass the message to the other clients
	for (i = 0; i < CLIENTS_MAX; i++) {
		// skip clients that have quitted the game
		if (game->clients[i].alive == NO)
			continue;

		addrClient->sin_port = game->clients[i].port;
		addrClient->sin_addr.s_addr = game->clients[i].iP;

		if (sendto(sockFD, outputBuffer, size, 0, (struct sockaddr *)addrClient, sizeof(struct sockaddr)) == -1)
			fprintf(stderr, "GAMEDISTRIBUTE: [1] Error sending a message.\n");
	}
}


void gameDistributeStatus(struct game *game, struct sockaddr_in *addrClient, unsigned char *outputBuffer, int sockFD) {

	int i, j, k;

	// header
	j = 2;
	outputBuffer[j++] = 'N';

	// the number of entries
	outputBuffer[j++] = CLIENTS_MAX;

	// ready? (these could be embedded into the name strings)
	for (i = 0; i < CLIENTS_MAX; i++) {
		if (game->clients[i].alive == NO)
			outputBuffer[j++] = NO;
		else
			outputBuffer[j++] = game->clients[i].ready;
	}

	// plane models
	for (i = 0; i < CLIENTS_MAX; i++)
		outputBuffer[j++] = game->clients[i].planeModel;

	// RGB values
	for (i = 0; i < CLIENTS_MAX; i++) {
		outputBuffer[j++] = game->clients[i].colorR;
		outputBuffer[j++] = game->clients[i].colorG;
		outputBuffer[j++] = game->clients[i].colorB;
	}

	// names
	for (i = 0; i < CLIENTS_MAX; i++) {
		k = 0;
		if (game->clients[i].alive == YES) {
			while (game->clients[i].name[k] != 0)
				outputBuffer[j++] = game->clients[i].name[k++];
		}
		outputBuffer[j++] = 0;
	}

	// the current level
	outputBuffer[j++] = game->level;

	// multiplayer mode
	outputBuffer[j++] = game->playMode;

	// the state bits
	outputBuffer[j++] = game->stateBits >> 24;
	outputBuffer[j++] = game->stateBits >> 16;
	outputBuffer[j++] = game->stateBits >> 8;
	outputBuffer[j++] = game->stateBits;

	// distribute the data
	gameDistribute(game, addrClient, outputBuffer, sockFD, j);
}


void gameDistributeGameStatistics(struct game *game, struct sockaddr_in *addrClient, unsigned char *outputBuffer, int sockFD) {

	int i, j;

	// header
	j = 2;
	outputBuffer[j++] = 's';

	// the number of entries
	outputBuffer[j++] = CLIENTS_MAX;

	// kills
	for (i = 0; i < CLIENTS_MAX; i++)
		outputBuffer[j++] = game->clients[i].kills;

	// distribute the data
	gameDistribute(game, addrClient, outputBuffer, sockFD, j);
}


void sigHandler(int sig) {

	fprintf(stderr, "SIGHANDLER: Saving the stats...\n");

	statsSave(&game);

	// make the program exit
	quit = YES;
}


void noIncomingTraffic(struct game *game, struct sockaddr_in *addrClient, unsigned char *outputBuffer, int sockFD) {

	struct timespec req, rem;
	struct timeval tv;
	long long ms, dt;
	int i;

	// sleep for 5ms
	req.tv_sec = 0;
	req.tv_nsec = 5000000;

	nanosleep(&req, &rem);

	// get the current time
	gettimeofday(&tv, NULL);

	// -> ms
	ms = ((long long)tv.tv_sec * 1000) + ((long long)tv.tv_usec / 1000);

	// do we need to ping the clients?
	for (i = 0; i < CLIENTS_MAX; i++) {
		// skip clients that have quitted the game
		if (game->clients[i].alive == NO)
			continue;

		dt = ms - game->clients[i].lastPingMS;

		if (dt > 1000) {
			// last ping was more than a second ago
			game->clients[i].lastPingMS = ms;

			// timeout the client?
			if (game->clients[i].pingsSent >= clientTimeoutSeconds) {
				// yes!
#if defined(DEBUG_MESSAGES)
				fprintf(stderr, "NOINCOMINGTRAFFIC: %s (%d) timed out.\n", game->clients[i].name, i);
#endif
				// send this information to all clients
				outputBuffer[2] = 'X';
				outputBuffer[3] = i;

				// distribute the data
				gameDistribute(game, addrClient, outputBuffer, sockFD, 4);

				// kill the client
				clientQuit(game, i);
			}
			else {
				// no, send a ping
				game->clients[i].pingsSent++;

				// set the size
				outputBuffer[0] = 3 >> 8;
				outputBuffer[1] = 3 & 0xFF;
				outputBuffer[2] = 'p';

				// fix the structure
				addrClient->sin_port = game->clients[i].port;
				addrClient->sin_addr.s_addr = game->clients[i].iP;

				// send
				if (sendto(sockFD, outputBuffer, 3, 0, (struct sockaddr *)addrClient, sizeof(struct sockaddr)) == -1)
					fprintf(stderr, "NOINCOMINGTRAFFIC: [1] Error sending a message.\n");
			}
		}
	}
}


int readPlaySettings(int clientID, unsigned char *buffer, int i) {

	// plane model
	game.clients[clientID].planeModel = buffer[i++];

	// RGB
	game.clients[clientID].colorR = buffer[i++];
	game.clients[clientID].colorG = buffer[i++];
	game.clients[clientID].colorB = buffer[i++];

	// level
	game.level = buffer[i++];

	// multiplayer mode
	game.playMode = buffer[i++];

	// configuration bits
	game.stateBits = (buffer[i+0] << 24) | (buffer[i+1] << 16) | (buffer[i+2] << 8) | buffer[i+3];
	i += 4;

	return i;
}


int main(int argc, char *argv[]) {

	struct sockaddr_in addrServer; // my address information
	struct sockaddr_in addrClient; // connector's address information
	unsigned char inputBuffer[INPUT_BUFFER_SIZE];
	unsigned char outputBuffer[INPUT_BUFFER_SIZE];
	unsigned long int clientIP;
	int packetSize, sockFD;
	socklen_t addrSize;
	int i, j, k, m, clientValid, clientID;

	// exit?
	i = NO;

	// parse the client timeout
	if (argc > 1)
		clientTimeoutSeconds = atoi(argv[1]);

	if (argc != 2 || clientTimeoutSeconds <= 0)
		i = YES;

	if (i == YES) {
		fprintf(stderr, "USAGE: %s <TIMEOUT SECONDS>\n", argv[0]);
		return 1;
	}

	// set the signal handlers
	signal(SIGINT, sigHandler);
	signal(SIGTERM, sigHandler);

	// reset the game (no clients, etc.)
	gameReset(&game);

	// import the stats
	statsLoad(&game);

	// create a socket
	if ((sockFD = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		fprintf(stderr, "MAIN: Error creating the socket.\n");
		exit(1);
	}

	// set the socket to be non-blocking
	i = 1;
	ioctl(sockFD, FIONBIO, &i);

	addrServer.sin_family = AF_INET;          // host byte order
	addrServer.sin_port = htons(SERVER_PORT); // short, network byte order
	addrServer.sin_addr.s_addr = INADDR_ANY;  // automatically fill with my IP

	if (bind(sockFD, (struct sockaddr *)&addrServer, sizeof(struct sockaddr)) == -1) {
		fprintf(stderr, "MAIN: Error binding the socket.\n");
		exit(1);
	}

	// print my info
	{
		struct hostent *serverInfo;
		char hostname[256];

		// get my hostname
		gethostname(hostname, 256);

		// get the corresponding IP
		serverInfo = gethostbyname(hostname);

		if (serverInfo != NULL) {
			addrServer.sin_addr.s_addr = *((unsigned long *)(serverInfo->h_addr_list[0]));
			fprintf(stderr, "MAIN: Server running at %s:%d...\n", inet_ntoa(addrServer.sin_addr), ntohs(addrServer.sin_port));
		}
		else
			fprintf(stderr, "MAIN: Server running at %s...\n", hostname);
	}

	// the main loop
	while (1) {
		// exit?
		if (quit == YES)
			break;

		// incoming traffic?
		addrSize = sizeof(struct sockaddr);
		packetSize = recvfrom(sockFD, inputBuffer, INPUT_BUFFER_SIZE, 0, (struct sockaddr *)&addrClient, &addrSize);

		if (packetSize <= 0) {
			// no incoming traffic -> ping?
			noIncomingTraffic(&game, &addrClient, outputBuffer, sockFD);
			continue;
		}

		// TODO: move the following code into its own function

		clientIP = addrClient.sin_addr.s_addr;
		clientValid = NO;

		// client validity check
		for (i = 0; i < CLIENTS_MAX; i++) {
			if (game.clients[i].alive == YES && game.clients[i].iP == clientIP) {
				clientValid = YES;
				clientID = i;
				break;
			}
		}

		/*
#if defined(DEBUG_MESSAGES)
		fprintf(stderr, "MAIN: Packet from %s:%d.\n", inet_ntoa(addrClient.sin_addr), ntohs(addrClient.sin_port));
#endif
		*/

		// is the packet size ok?
		if (packetSize < 3) {
			fprintf(stderr, "MAIN: The packet (%d bytes) is too small!\n", packetSize);
			continue;
		}

		// is the embedded size ok?
		i = (inputBuffer[0] << 8) | inputBuffer[1];
		if (i != packetSize) {
			fprintf(stderr, "MAIN: The actual packet size (%d bytes) and the embedded packet size (%d bytes) don't match!\n", packetSize, i);
			continue;
		}

		///////////////////////////////////////////////////////////////////////////////////////////////////
		// COMMON MESSAGES
		///////////////////////////////////////////////////////////////////////////////////////////////////

		if (clientValid == YES) {
			if (inputBuffer[2] == 'u') {
				// scroll up
				game.clients[clientID].statisticsScrollY--;
				if (game.clients[clientID].statisticsScrollY < 0)
					game.clients[clientID].statisticsScrollY = 0;

				// send the results
				gameSendStatistics(&game, &addrClient, outputBuffer, sockFD, clientID);

				continue;
			}
			else if (inputBuffer[2] == 'd') {
				// scroll down
				game.clients[clientID].statisticsScrollY++;

				// send the results
				gameSendStatistics(&game, &addrClient, outputBuffer, sockFD, clientID);

				continue;
			}
			else if (inputBuffer[2] == 'r') {
				// refresh

				// send the results
				gameSendStatistics(&game, &addrClient, outputBuffer, sockFD, clientID);

				continue;
			}
			else if (inputBuffer[2] == 'p') {
				// ping

				// reset the ping counter
				game.clients[clientID].pingsSent = 0;

				continue;
			}
			else if (inputBuffer[2] == 'm') {
				readPlaySettings(clientID, inputBuffer, 3);

				// send all clients the client status list
				gameDistributeStatus(&game, &addrClient, outputBuffer, sockFD);

				continue;
			}
		}

		if (game.mode == GAME_MODE_TAKING_CLIENTS) {
			///////////////////////////////////////////////////////////////////////////////////////////////////
			// GAME_MODE_TAKING_CLIENTS
			///////////////////////////////////////////////////////////////////////////////////////////////////

			// a new client?
			if (inputBuffer[2] == 'J') {
				// yes! return the client ID number, if possible
				for (i = 0; i < CLIENTS_MAX; i++) {
					if (game.clients[i].iP == 0) {
						game.clients[i].iP = clientIP;
						game.clients[i].port = addrClient.sin_port;
						game.clients[i].alive = YES;
						game.clients[i].pingsSent = 0;

						// copy the name
						m = 3+10;

						k = inputBuffer[m++];
						for (j = 0; j < k; j++)
							game.clients[i].name[j] = inputBuffer[m++];
						game.clients[i].name[j] = 0;

						// copy the MAC address
						for (j = 0; j < 6; j++)
							game.clients[i].macAddress[j] = inputBuffer[m++];

						game.clientsN++;

						// add the MAC address
						statsMACAdd(&game, &(game.clients[i]));

						// find the stats table
						statsGetTable(&game, &(game.clients[i]));

#if defined(DEBUG_MESSAGES)
						{
							unsigned char *mac = game.clients[i].macAddress;

							fprintf(stderr, "MAIN: Got a new client %s (%.2X%.2X%.2X%.2X%.2X%.2X) %s:%d (%d).\n", game.clients[i].name, mac[0], mac[1], mac[2], mac[3], mac[4], mac[5], inet_ntoa(addrClient.sin_addr), ntohs(game.clients[i].port), i);
						}
#endif
					}
					if (game.clients[i].iP == clientIP) {
						// input the client configuration
						readPlaySettings(i, inputBuffer, 3);

						outputBuffer[0] = 0; // the size of the message byte 1.
						outputBuffer[1] = 4; //                              2.
						outputBuffer[2] = 'W';
						outputBuffer[3] = i;

						// send the client a welcome!
						if (sendto(sockFD, outputBuffer, 4, 0, (struct sockaddr *)&addrClient, sizeof(struct sockaddr)) == -1)
							fprintf(stderr, "MAIN: [1] Error sending a message.\n");

						// send the statistics
						gameSendStatistics(&game, &addrClient, outputBuffer, sockFD, i);

						// send all clients the client status list
						gameDistributeStatus(&game, &addrClient, outputBuffer, sockFD);

						break;
					}
				}

				// are we out of client slots?
				if (i == CLIENTS_MAX) {
					// yes, send an apology
					outputBuffer[0] = 0; // the size of the message byte 1.
					outputBuffer[1] = 3; //                              2.
					outputBuffer[2] = 'S';

					if (sendto(sockFD, outputBuffer, 3, 0, (struct sockaddr *)&addrClient, sizeof(struct sockaddr)) == -1)
						fprintf(stderr, "MAIN: [2] Error sending a message.\n");
				}
			}
			// begin the game?
			else if (inputBuffer[2] == 'B') {
				if (clientValid == YES) {
					// mark as ready!
					game.clients[clientID].ready = YES;

					// parse the client's settings
					readPlaySettings(clientID, inputBuffer, 3);

					// send all clients the client status list
					gameDistributeStatus(&game, &addrClient, outputBuffer, sockFD);

					// all ready?
					j = 0;
					for (i = 0; i < CLIENTS_MAX; i++) {
						if (game.clients[i].alive == NO)
							continue;
						if (game.clients[i].ready == NO)
							break;
						j++;
					}

					if (i == CLIENTS_MAX && j > 1) {
						// YES!

						// increase the play count
						for (k = 0; k < CLIENTS_MAX; k++) {
							if (game.clients[k].alive == YES)
								game.clients[k].stats->plays++;
						}

						// inform the clients
						j = 2;

						outputBuffer[j++] = 'B';
						outputBuffer[j++] = CLIENTS_MAX;

						// write the IPs and the ports
						for (k = 0; k < CLIENTS_MAX; k++) {
							// alive
							outputBuffer[j++] = game.clients[k].alive;

							if (game.clients[k].alive == YES) {
								// IP
								outputBuffer[j++] = game.clients[k].iP >> 24;
								outputBuffer[j++] = game.clients[k].iP >> 16;
								outputBuffer[j++] = game.clients[k].iP >> 8;
								outputBuffer[j++] = game.clients[k].iP;

								// port
								outputBuffer[j++] = game.clients[k].port >> 8;
								outputBuffer[j++] = game.clients[k].port;
							}
						}

						// send the message to all clients
						gameDistribute(&game, &addrClient, outputBuffer, sockFD, j);

						// change the game mode
						game.mode = GAME_MODE_GAME;

#if defined(DEBUG_MESSAGES)
						fprintf(stderr, "MAIN: A game has begun!\n");
#endif
					}
				}
			}
			else {
#if defined(DEBUG_MESSAGES)
				fprintf(stderr, "MAIN: Got an unknown message from %s.\n", inet_ntoa(addrClient.sin_addr));
#endif
			}
		}

		// invalid clients don't get any further
		if (clientValid == NO)
			continue;

		if (game.mode == GAME_MODE_GAME) {
			///////////////////////////////////////////////////////////////////////////////////////////////////
			// GAME_MODE_GAME
			///////////////////////////////////////////////////////////////////////////////////////////////////

			// TODO: analyze the message

			i = 2;

			switch (inputBuffer[i++]) {
			case 'K':
				// KILL
				j = inputBuffer[i];

#if defined(DEBUG_MESSAGES)
				fprintf(stderr, "MAIN: %s (%d) killed %s (%d).\n", game.clients[j].name, j, game.clients[clientID].name, clientID);
#endif

				// give a minus point for suicide
				if (j != clientID)
					game.clients[j].kills++;
				else
					game.clients[j].kills--;

				if (game.clients[j].kills < 0)
					game.clients[j].kills = 0;

				// send everybody the updated statistics
				gameDistributeGameStatistics(&game, &addrClient, outputBuffer, sockFD);

				// catalog the kill
				if (j != clientID)
					statsAddKill(&(game.clients[j]), &(game.clients[clientID]));

				// send the statistics
				gameSendStatistics(&game, &addrClient, outputBuffer, sockFD, j);

				if (j != clientID)
					gameSendStatistics(&game, &addrClient, outputBuffer, sockFD, clientID);

				break;

			case 'D':
				// DEATH - A CLIENT DROPPED OUT

#if defined(DEBUG_MESSAGES)
				fprintf(stderr, "MAIN: %s (%d) has quit the game.\n", game.clients[clientID].name, clientID);
#endif
				// kill the client
				clientQuit(&game, clientID);

				if (game.clientsN > 0) {
					// send this information to all the other clients
					outputBuffer[2] = 'X';
					outputBuffer[3] = clientID;

					// distribute the data
					gameDistribute(&game, &addrClient, outputBuffer, sockFD, 4);
				}

				break;

			case 'Q':
				// QUIT SESSION

				// reset the game (no clients, etc.)
				gameReset(&game);

#if defined(DEBUG_MESSAGES)
				fprintf(stderr, "MAIN: Quit session requested... Resetting...\n");
#endif

				break;
			}

			/*
			// if the first event in the message is '!', then the server should analyze the message
			if (inputBuffer[2] == '!') {
				// yes! analyze


			}

			// pass the message to the other clients
			for (i = 0; i < game.clientsN; i++) {
				// skip the sender itself
				if (i == clientID)
					continue;

				addrClient.sin_port = game.clients[i].port;
				addrClient.sin_addr.s_addr = game.clients[i].iP;

				if (sendto(sockFD, inputBuffer, packetSize, 0, (struct sockaddr *)&addrClient, sizeof(struct sockaddr)) == -1) {
					fprintf(stderr, "MAIN: [4] Error sending a message.\n");
				}
			}
			*/
		}

		// when we reach here, the packet we got from a valid client has been processed
	}

	close(sockFD);

	return 0;
} 
