
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "defines.h"
#include "main.h"
#include "game.h"
#include "stats.h"


// the stats database file name
#define STATS_FILE_NAME "STATS.DB"


struct vs *statsGetVS(struct client *source, struct client *target) {

	struct vs *vs;
	int i;

	// does the record exist?
	vs = source->stats->vs;
	while (vs != NULL) {
		if (vs->macAddress[0] == target->macAddress[0] &&
				vs->macAddress[1] == target->macAddress[1] &&
				vs->macAddress[2] == target->macAddress[2] &&
				vs->macAddress[3] == target->macAddress[3] &&
				vs->macAddress[4] == target->macAddress[4] &&
				vs->macAddress[5] == target->macAddress[5]) {
			// yes, got a match!
			return vs;
		}
		vs = vs->next;
	}

	// no match -> create the record
	vs = malloc(sizeof(struct vs));
	if (vs == NULL) {
		fprintf(stderr, "STATSGETVS: Out of memory error [1].\n");
		return NULL;
	}

	for (i = 0; i < 6; i++)
		vs->macAddress[i] = target->macAddress[i];

	vs->wins = 0;
	vs->losses = 0;
	vs->next = source->stats->vs;
	source->stats->vs = vs;

	return vs;
}


void statsAddKill(struct client *killer, struct client *killee) {

	struct vs *vsKiller, *vsKillee;

	vsKiller = statsGetVS(killer, killee);
	vsKillee = statsGetVS(killee, killer);

	vsKiller->wins++;
	vsKillee->losses++;
}


struct mac *statsMACGet(struct game *game, unsigned char *macAddress) {

	struct mac *mac;

	mac = game->macs;
	while (mac != NULL) {
		if (mac->macAddress[0] == macAddress[0] &&
				mac->macAddress[1] == macAddress[1] &&
				mac->macAddress[2] == macAddress[2] &&
				mac->macAddress[3] == macAddress[3] &&
				mac->macAddress[4] == macAddress[4] &&
				mac->macAddress[5] == macAddress[5]) {
			// yes, got a match!
			return mac;
		}
		mac = mac->next;
	}

	return NULL;
}


void statsMACAdd(struct game *game, struct client *client) {

	struct mac *mac;
	int i;

	// is the MAC address already in our list?
	mac = statsMACGet(game, client->macAddress);
	if (mac != NULL) {
		// yes, got a match! update the name and return
		strcpy(mac->name, client->name);
		return;
	}

	// no match -> add it to the list
	mac = malloc(sizeof(struct mac));
	if (mac == NULL) {
		fprintf(stderr, "STATSMACADD: Out of memory error [1].\n");
		return;
	}

	// copy MAC address
	for (i = 0; i < 6; i++)
		mac->macAddress[i] = client->macAddress[i];

	// copy name
	strcpy(mac->name, client->name);

	// add it to the list
	mac->next = game->macs;
	game->macs = mac;
}


void statsGetTable(struct game *game, struct client *client) {

	struct stats *stats;
	int i;

	// clear the stats pointer...
	client->stats = NULL;

	// is there a matching client?
	stats = game->stats;
	while (stats != NULL) {
		if (stats->macAddress[0] == client->macAddress[0] &&
				stats->macAddress[1] == client->macAddress[1] &&
				stats->macAddress[2] == client->macAddress[2] &&
				stats->macAddress[3] == client->macAddress[3] &&
				stats->macAddress[4] == client->macAddress[4] &&
				stats->macAddress[5] == client->macAddress[5]) {
			// yes, got a match!
			client->stats = stats;
			return;
		}

		stats = stats->next;
	}

	// no match, allocate a new one
	stats = malloc(sizeof(struct stats));
	if (stats == NULL) {
		fprintf(stderr, "STATSGETTABLE: Out of memory error [1].\n");
		return;
	}

	stats->plays = 0;
	stats->vs = NULL;
	for (i = 0; i < 6; i++)
		stats->macAddress[i] = client->macAddress[i];

	stats->next = game->stats;
	game->stats = stats;

	client->stats = stats;
}


void statsLoad(struct game *game) {

	struct stats *stats, **statsPtr = &(game->stats);
	struct mac *mac, **macPtr = &(game->macs);
	struct vs *vs, **vsPtr;
	int i, j, k, statsN, macsN;
	FILE *f;

	f = fopen(STATS_FILE_NAME, "rb");
	if (f == NULL) {
		fprintf(stderr, "STATSLOAD: Couldn't open file \"%s\" for reading.\n", STATS_FILE_NAME);
		return;
	}

	game->stats = NULL;
	game->macs = NULL;

	// the number of macs
	if (fscanf(f, "%d", &macsN) == EOF) {
		fclose(f);
		return;
	}

	// macs
	while (!feof(f) && macsN > 0) {
		mac = malloc(sizeof(struct mac));
		if (mac == NULL) {
			fprintf(stderr, "STATSLOAD: Out of memory error [3].\n");
			return;
		}

		mac->next = NULL;

		// MAC address
		for (i = 0; i < 6; i++) {
			fscanf(f, "%d", &j);
			mac->macAddress[i] = j;
		}

		// name
		fscanf(f, "%15s", mac->name);

		*macPtr = mac;
		macPtr = &(mac->next);

		macsN--;
	}

	// the number of stats
	if (fscanf(f, "%d", &statsN) == EOF) {
		fclose(f);
		return;
	}

	while (!feof(f) && statsN > 0) {
		stats = malloc(sizeof(struct stats));
		if (stats == NULL) {
			fprintf(stderr, "STATSLOAD: Out of memory error [1].\n");
			return;
		}

		stats->next = NULL;
		stats->vs = NULL;

		// MAC address
		for (i = 0; i < 6; i++) {
			fscanf(f, "%d", &j);
			stats->macAddress[i] = j;
		}

		*statsPtr = stats;
		statsPtr = &(stats->next);

		// the number of played games
		fscanf(f, "%d", &(stats->plays));

		// the number of vs items
		fscanf(f, "%d", &k);

		vsPtr = &(stats->vs);

		while (k > 0) {
			vs = malloc(sizeof(struct vs));
			if (vs == NULL) {
				fprintf(stderr, "STATSLOAD: Out of memory error [2].\n");
				return;
			}

			vs->next = NULL;

			*vsPtr = vs;
			vsPtr = &(vs->next);

			// MAC address
			for (i = 0; i < 6; i++) {
				fscanf(f, "%d", &j);
				vs->macAddress[i] = j;
			}

			// wins
			fscanf(f, "%d", &(vs->wins));

			// losses
			fscanf(f, "%d", &(vs->losses));

			k--;
		}

		statsN--;
	}

	fclose(f);
}


void statsSave(struct game *game) {

	struct stats *stats;
	struct mac *mac;
	struct vs *vs;
	FILE *f;
	int i;

	f = fopen(STATS_FILE_NAME, "wb");
	if (f == NULL) {
		fprintf(stderr, "STATSSAVE: Couldn't open file \"%s\" for writing.\n", STATS_FILE_NAME);
		return;
	}

	// count the macs
	i = 0;
	mac = game->macs;
	while (mac != NULL) {
		i++;
		mac = mac->next;
	}

	fprintf(f, "%d\n", i);

	mac = game->macs;
	while (mac != NULL) {
		// MAC address
		for (i = 0; i < 6; i++)
			fprintf(f, "%d ", mac->macAddress[i]);

		// name
		fprintf(f, "%s\n", mac->name);

		mac = mac->next;
	}

	// count the stats
	i = 0;
	stats = game->stats;
	while (stats != NULL) {
		i++;
		stats = stats->next;
	}

	// the number of stats
	fprintf(f, "%d\n", i);

	stats = game->stats;
	while (stats != NULL) {
		// MAC address
		for (i = 0; i < 6; i++)
			fprintf(f, "%d ", stats->macAddress[i]);
		fprintf(f, "\n");

		// the number of played games
		fprintf(f, "%d\n", stats->plays);

		// count the vs items
		i = 0;
		vs = stats->vs;
		while (vs != NULL) {
			i++;
			vs = vs->next;
		}

		// the number of vs items
		fprintf(f, "%d\n", i);

		// the vs items
		vs = stats->vs;
		while (vs != NULL) {
			// MAC address
			for (i = 0; i < 6; i++)
				fprintf(f, "%d ", vs->macAddress[i]);
			fprintf(f, "\n");

			// wins
			fprintf(f, "%d\n", vs->wins);

			// losses
			fprintf(f, "%d\n", vs->losses);

			vs = vs->next;
		}

		stats = stats->next;
	}

	fclose(f);
}
